// 008 u_cgi_helpers
// 27 may 2003

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_cgi_helpers;
  interface
    uses Classes;

    function f_cgi_exe_path: String;

    // -- environment handling

    function f_get_all_environment: String;
    function f_get_environment_value(p_key: pChar): String;
    function f_request_method: String;
    function f_content_length: Integer;

    function f_raw_request_string: String;
    function f_convert_special_characters(p_string: String): String;
    // function f_convert_cgi_question_string(p_cgi_string: String): String;
    function f_c_parameter_list: tStringList;

  implementation
    uses SysUtils, Windows;

    // -- server pathes

    function f_cgi_exe_path: String;
      var l_size: Integer;
          l_cgi_exe_name_array: array[0..MAX_PATH] of Char;
          l_cgi_exe_name: String;
      begin
        l_size:= Windows.GetModuleFileName(hInstance, l_cgi_exe_name_array, MAX_PATH);
        SetString(l_cgi_exe_name, l_cgi_exe_name_array, l_size);
        Result:= ExtractFilePath(l_cgi_exe_name);
      end; // f_cgi_exe_path

    // -- environment variables

    function f_get_all_environment: String;
        // -- get all the environment variables
      var l_pchar: PChar;
          l_environment_line: ShortString;
      begin
        l_pchar:= GetEnvironmentStrings;
        Result:= '';

        while l_pchar^<> #0 do
        begin
          l_environment_line:= StrPas(l_pchar);
          Result:= Result+ l_environment_line+ chr(13)+chr(10)+ '    ';
          Inc(l_pchar, StrLen(l_pchar)+ 1)
        end; // wile

        FreeEnvironmentStrings(l_pchar);
      end; // f_get_all_environment

    function f_get_environment_value(p_key: pChar): String;
      const k_max= 255;
      var l_environment_value: String[k_max];
          l_size: Integer;
      begin
        l_size:= GetEnvironmentVariable(p_key, @ l_environment_value[1], k_max);
        l_environment_value[0]:= chr(l_size);
        Result:= l_environment_value;
      end; // f_get_environment_value

    // -- request handling

    function f_request_method: String;
      begin
        Result:= f_get_environment_value('REQUEST_METHOD');
      end; // f_request_method

    function f_content_length: Integer;
      var l_content_length: String;
      begin
        l_content_length:= f_get_environment_value('CONTENT_LENGTH');
        if l_content_length<> ''
          then Result:= StrToInt(l_content_length)
          else Result:= 0;
      end; // f_content_length

    var // avoid reading twice the parameters
        s_raw_request_string: String= '';

    function f_raw_request_string: String;
      // -- CAUTION: only read values ONCE (else hangs)
      var l_request_method: String;
          l_content_length: Integer;
          l_index: Integer;
      begin
        if s_raw_request_string<> ''
          then Result:= s_raw_request_string
          else begin
              l_request_method:= f_request_method;
              l_content_length:= f_content_length;

              // -- now read the question
              if (l_request_method= 'POST') and (l_content_length> 0)
                then begin
                    // -- +1 for last & added by us
                    SetLength(Result, l_content_length);

                    for l_index:= 1 to l_content_length do
                      Read(Result[l_index]);

                    s_raw_request_string:= Result;
                  end
                else Result:= '';
            end;
      end; // f_raw_request_string

    function f_convert_special_characters(p_string: String): String;
        // -- convert
        // --   + to space
        // --   $xx to chr($xx)
      var l_hex_string: ShortString;
          l_index: Integer;
      begin
        l_index:= 1;
        Result:= '';

        while l_index<= Length(p_string) do
        begin
          if p_string[l_index]= '+'
            then // -- change the + back to spaces
                 Result:= Result+ ' '
            else
              if p_string[l_index]= '%'
                then
                  begin
                    // -- change specials %26 into &
                    l_hex_string:= '$00';
                    l_hex_string[2]:= p_string[l_index+ 1];
                    l_hex_string[3]:= p_string[l_index+ 2];
                    Result:= Result+ Chr(StrToInt(l_hex_string));
                    Inc(l_index, 2);
                  end
                else Result:= Result+ p_string[l_index];

          Inc(l_index);
        end; // while
      end; // f_convert_special_characters

    function f_c_parameter_list: tStringList;
      var l_raw_request_string, l_request_string: String;
          l_index: Integer;
          l_parameter: String;
      begin
        Result:= tStringList.Create;

        l_raw_request_string:= f_raw_request_string;
        l_request_string:= f_convert_special_characters(l_raw_request_string);

        // -- format: xxx=yyy&xxx=yyy
        l_parameter:= '';
        for l_index:= 1 to Length(l_request_string) do
        begin
          if l_request_string[l_index]= '&'
            then begin
                Result.Add(l_parameter);
                l_parameter:= '';
              end
            else l_parameter:= l_parameter+ l_request_string[l_index];
        end; // for

        // -- add the last one
        if l_parameter<> ''
          then Result.Add(l_parameter);
      end; // f_c_parameter_list

    begin
    end.

