// 003 u_c_cgi_request_evaluator
// 02 jan 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

// -- parse the expression, and check whether a text satisfies the condition

(*$r+*)

unit u_c_cgi_request_evaluator;
  interface
    uses Classes, u_c_cgi_basic_object
        , u_c_cgi_text_searcher
      ;

    type t_error_type= (e_unknown_error
                         , e_missing_ending_quote
                         , e_illegal_character
                         , e_missing_closing_parenthesis
                         , e_expects_literal
                         , e_stop_before_end
                         );

         c_request_evaluator= class(c_basic_object)
                                m_trace_tokenize: Boolean;
                                m_trace_evaluation: Boolean;

                                // -- the original request
                                m_request: String;
                                // -- the array of literals
                                m_c_request_identifiers: tStringList;
                                m_token_count: Integer;

                                // -- the array of tokens: negative for NOT AND OR  ( )
                                // -- positive for literals
                                m_oa_tokens: Array Of Integer;
                                // -- for error positionning
                                m_oa_token_positions: Array Of Integer;
                                m_oa_token_lengths: Array Of Integer;

                                m_error_index, m_error_length: Integer;
                                m_error_type: t_error_type;
                                m_error_message: String;
                                // -- debug
                                m_error_token_index: Integer;

                                Constructor create_request_evaluator(p_name: String);

                                function f_c_self: c_request_evaluator;
                                function f_tokenize_request(p_request: String): Boolean;
                                procedure display_arrays;
                                function f_display_token_type(p_token_type: Integer): String;
                                procedure redisplay_request;

                                function f_check_syntax: Boolean;

                                function f_evaluate_request(p_c_text_searcher: c_text_searcher): Boolean;

                                Destructor Destroy; Override;
                              end; // c_request_evaluator

    function f_error_message(p_error_type: t_error_type): String;

  implementation
    uses SysUtils
        , u_cgi_strings, u_cgi_display
        ;

    function f_error_message(p_error_type: t_error_type): String;
      begin
        case p_error_type of
          e_missing_ending_quote : Result:= 'missing ending "';
          e_illegal_character : Result:= 'illegal_character';
          e_missing_closing_parenthesis : Result:= 'missing )';
          e_expects_literal : Result:= 'expects name or "string"';
          e_stop_before_end : Result:= 'stopped before end '
              + '(maybe missing AND OR NOT, missing " " for strings with spaces)';
          else
            result:= 'bug_unknown';
        end; // case p_error_type
      end; // f_error_message

    // -- c_request_evaluator

    Constructor c_request_evaluator.create_request_evaluator(p_name: String);
      begin
        Inherited ;
      end; // create_request_evaluator

    function c_request_evaluator.f_c_self: c_request_evaluator;
      begin
        Result:= Self;
      end; // f_c_self

    const k_token_NOT= -1;
          k_token_AND= -2;
          k_token_OR= -3;
          k_token_opening_parenthesis= -4;
          k_token_closing_parenthesis= -5;
          k_token_end= -6;

    function c_request_evaluator.f_tokenize_request(p_request: String): Boolean;
      const k_start_letters= ['a'..'z', 'A'..'Z'];
            k_letters= k_start_letters+ ['_', '0'..'9'];
      var l_request_length: Integer;
          l_request_index: Integer;
          l_token_max: Integer;

      procedure trace_tokenize(p_text: String);
        begin
          if m_trace_tokenize
            then display(p_text);
        end; // trace_tokenize

      procedure set_lexical_error(p_error_type: t_error_type; p_error_index, p_error_length: Integer);
        begin
          m_error_type:= p_error_type;
          m_error_index:= p_error_index;
          m_error_length:= p_error_length;
        end; // set_lexical_error

      procedure set_arrays_capacity(p_capacity: Integer);
        begin
          trace_tokenize('= buddy '+ IntToStr(p_capacity));
          l_token_max:= p_capacity;
          SetLength(m_oa_tokens, p_capacity);
          SetLength(m_oa_token_positions, p_capacity);
          SetLength(m_oa_token_lengths, p_capacity);
        end; // set_arrays_capacity

      procedure add_token(p_token, p_start_index, p_length: Integer);
        begin
          trace_tokenize('> add '+ IntToStr(p_token));
          if m_token_count>= l_token_max
            then // -- buddy: double size
                 set_arrays_capacity(2* l_token_max);

          m_oa_tokens[m_token_count]:= p_token;
          m_oa_token_positions[m_token_count]:= p_start_index;
          m_oa_token_lengths[m_token_count]:= p_length;

          Inc(m_token_count);
          trace_tokenize('< add ');
        end; // add_token

      procedure analyze_identifier;
        var l_start_index, l_identifier_length: Integer;
            l_identifier: String;
            // l_index: Integer;
            l_lowercase_identifier: String;
            l_the_token: Integer;
        begin
          trace_tokenize(Format('%3d %3d ', [l_request_index, l_request_length]));
          l_start_index:= l_request_index;
          while (l_request_index<= l_request_length)
              and (m_request[l_request_index] in k_letters) do
            Inc(l_request_index);

          l_identifier_length:= l_request_index- l_start_index;
          SetLength(l_identifier, l_identifier_length);
          Move(m_request[l_start_index], l_identifier[1], l_identifier_length);

          l_lowercase_identifier:= LowerCase(l_identifier);

          // -- check whether this is AND OR NOT
          if l_lowercase_identifier= 'and'
            then l_the_token:= k_token_AND
              else
                if l_lowercase_identifier= 'or'
                  then l_the_token:= k_token_OR
                    else
                      if l_lowercase_identifier= 'not'
                        then l_the_token:= k_token_NOT
                        else // -- the identifier
                             // l_the_token:= m_c_request_identifiers.Add(l_lowercase_identifier);
                             l_the_token:= m_c_request_identifiers.Add(l_identifier);

          add_token(l_the_token, l_start_index, l_identifier_length);
        end; // analyze_identifier

      procedure analyze_litteral_string;
        var l_start_index, l_identifier_length: Integer;
            l_identifier, l_lowercase_identifier: String;
            l_index: Integer;
        begin
          // -- skip "
          Inc(l_request_index);

          l_start_index:= l_request_index;
          while (l_request_index<= l_request_length)
              and (m_request[l_request_index]<> '"') do
            Inc(l_request_index);

          l_identifier_length:= l_request_index- l_start_index;
          SetLength(l_identifier, l_identifier_length);
          Move(m_request[l_start_index], l_identifier[1], l_identifier_length);

          l_index:= m_c_request_identifiers.Add(l_identifier);

          add_token(l_index, l_start_index, l_identifier_length);

          if l_request_index>= l_request_length
            then set_lexical_error(e_missing_ending_quote, l_request_index- 1, 1);

          display(Format('%3d %3d ', [l_request_index, l_request_length]));  
          // -- skip "
          Inc(l_request_index);
        end; // analyze_litteral_string

      begin // f_tokenize_request
        trace_tokenize('> f_tokenize ');

        // -- create the identifier list
        m_c_request_identifiers:= tStringList.Create;
        // -- do NOT use index 0 (debug)
        m_c_request_identifiers.Add('');

        // -- initial capacity: 16
        set_arrays_capacity(16);

        m_request:= p_request+ '.';
        // trace_tokenize('|'+ m_request+ '| '+ IntToStr(Length(m_request)));

        l_request_length:= Length(m_request);

        l_request_index:= 1;
        m_token_count:= 0;
        m_error_index:= 0;

        while (l_request_index<= l_request_length) and (m_error_index= 0) do
        begin
          case m_request[l_request_index] of
            'a'..'z', 'A'..'Z',
              '0'..'9' : analyze_identifier;
            '"' : analyze_litteral_string;
            ' ' : Inc(l_request_index);
            '(' : begin
                    add_token(k_token_opening_parenthesis, l_request_index, 1);
                    Inc(l_request_index);
                  end;
            ')' : begin
                    add_token(k_token_closing_parenthesis, l_request_index, 1);
                    Inc(l_request_index);
                  end;
            '.' : begin
                    add_token(k_token_end, l_request_index, 1);
                    Inc(l_request_index);
                  end;
            else
              set_lexical_error(e_illegal_character, l_request_index, 1);
              Break;
          end; // case
        end; // while

        display('err '+ IntToStr(m_error_index));
        Result:= m_error_index= 0;

        trace_tokenize('< f_tokenize');
      end; // f_tokenize_request

    function c_request_evaluator.f_display_token_type(p_token_type: Integer): String;
      begin
        case p_token_type  of
          k_token_AND : Result:= 'AND';
          k_token_OR : Result:= 'OR';
          k_token_NOT : Result:= 'NOT';
          k_token_opening_parenthesis : Result:= '(';
          k_token_closing_parenthesis : Result:= ')';
          k_token_end : Result:= '.';
          else
            Result:= m_c_request_identifiers[p_token_type];
        end; // case
      end; // f_display_token_type

    procedure c_request_evaluator.display_arrays;
      // var l_index: Integer;
      begin
(*
        display('> display_arrays');

        display_strings('lit ', m_c_request_identifiers);

        For l_index:= 0 to m_token_count- 1 do
          display(Format('%3d  [%3d]', [l_index, m_oa_tokens[l_index]]));

        display('< display_arrays');
*)
      end; // display_arrays

    procedure c_request_evaluator.redisplay_request;
      var l_index: Integer;
          l_request: String;
      begin
        display('> redisplay');
        l_request:= '';

        For l_index:= 0 to m_token_count- 1 do
          if m_oa_tokens[l_index]<> 0
            then l_request:= l_request+ f_display_token_type(m_oa_tokens[l_index])+ ' ';

        display(l_request);
        display('< redisplay');
      end; // redisplay_request

    function c_request_evaluator.f_check_syntax: Boolean;
        // -- simply check the syntax
      type t_symbol_type= Integer;
      var // -- the position in the tokenized request
          l_token_index: Integer;
          // -- the current symbol
          l_symbol_type: t_symbol_type;
          l_error_count: Integer;

      procedure trace_evaluation(p_text: String);
        begin
          if m_trace_evaluation
            then display(p_text);
        end; // trace_evaluation

      procedure read_next_symbol;
        begin
          l_symbol_type:= m_oa_tokens[l_token_index];
          trace_evaluation('next_tok, ix= '+ IntToStr(l_token_index)+ ' type= '+ IntToStr(l_symbol_type)
            + ' '+ f_display_token_type(l_symbol_type));

          Inc(l_token_index);
        end; // read_next_symbol

      procedure add_syntax_error(p_error_type: t_error_type);
        begin
          Inc(l_error_count);
          m_error_type:= p_error_type;

          // -- compute the index of the error
          m_error_index:= m_oa_token_positions[l_token_index];
          m_error_length:= m_oa_token_lengths[l_token_index];
          m_error_token_index:= l_token_index;
        end; // add_syntax_error

      procedure evaluate_expression;

        procedure evaluate_term;

          procedure evaluate_factor;
            var l_identifier: String;
            begin
              trace_evaluation('> factor ix= '+ IntToStr(l_token_index)+ ' val= '+ f_display_token_type(l_symbol_type));

              case l_symbol_type of
                k_token_opening_parenthesis :
                  begin
                    trace_evaluation('parent');
                    // -- skip (
                    read_next_symbol;

                    evaluate_expression;

                    if l_symbol_type= k_token_closing_parenthesis
                      then read_next_symbol
                      else add_syntax_error(e_missing_closing_parenthesis);
                  end;

                k_token_NOT :
                  begin
                    trace_evaluation('NOT');
                    read_next_symbol;

                    evaluate_factor;
                    // -- do NOT read next symbol: was read by compile_expression
                  end;

                else
                  if l_symbol_type>= 0
                    then begin
                        trace_evaluation('get id '+ IntToStr(l_symbol_type));

                        l_identifier:= m_c_request_identifiers[l_symbol_type];
                        trace_evaluation('  '+ l_identifier);

                        // -- look if the symbol is in the text
                        // Result:= p_c_text_searcher.f_found_string(l_identifier);
                        read_next_symbol;
                      end
                    else begin
                        // -- avoid warnings
                        // Result:= False;
                        add_syntax_error(e_expects_literal); // 'display_evaluation_error('attend fact: k, var, (: '+ IntToStr(l_symbol_type));
                      end;
              end; // case

              trace_evaluation('< factor ');
            end; // evaluate_factor

          begin // evaluate_term
            trace_evaluation('> term '+ IntToStr(l_token_index));

            evaluate_factor;

            while (l_error_count= 0) and (l_symbol_type= k_token_AND) do
            begin
              read_next_symbol;
              evaluate_factor;
            end; // while l_symbol_type

            trace_evaluation('< term ');
          end; // evaluate_term

        begin // evaluate_expression
          trace_evaluation('> expression '+ IntToStr(l_token_index));

          evaluate_term;

          while (l_error_count= 0) and (l_symbol_type= k_token_OR) do
          begin
            read_next_symbol;

            evaluate_term;
          end; // while

          trace_evaluation('< expression ');
        end; // evaluate_expression

      begin // f_check_syntax
        l_error_count:= 0;

        l_token_index:= 0;
        read_next_symbol;

        evaluate_expression;

        if l_token_index< m_token_count
          then begin
              Dec(l_token_index);
              add_syntax_error(e_stop_before_end);
            end;
        // display(Format('%3d %3d', [l_token_index, m_token_count]));

        Result:= l_error_count= 0;

        // -- debug
        if not Result
            and ((m_error_index= 0) or (m_error_length= 0))
          then begin
              // -- the error was after the . => get back
              Dec(l_token_index);
              m_error_index:= m_oa_token_positions[l_token_index];
              m_error_length:= m_oa_token_lengths[l_token_index];

              // m_error_message:= 'Z';
            end;

        if not Result
          then m_error_message:= f_error_message(m_error_type);
      end; // f_check_syntax

    function c_request_evaluator.f_evaluate_request(p_c_text_searcher: c_text_searcher): boolean;
        // -- there should be no syntax error here
      type t_symbol_type= Integer;
      var // -- the position in the tokenized request
          l_token_index: Integer;
          // -- the current symbol
          l_symbol_type: t_symbol_type;
          // l_error_count: Integer;
          // l_first_error_token_index: Integer;

      procedure trace_evaluation(p_text: String);
        begin
          if m_trace_evaluation
            then display(p_text);
        end; // trace_evaluation

      procedure read_next_symbol;
        begin
          l_symbol_type:= m_oa_tokens[l_token_index];
          trace_evaluation('next_tok, ix= '+ IntToStr(l_token_index)+ ' type= '+ IntToStr(l_symbol_type)
            + ' '+ f_display_token_type(l_symbol_type));

          Inc(l_token_index);
        end; // read_next_symbol

      procedure display_evaluation_error(p_error: String);
        begin
(*
          trace_evaluation(m_request);
          trace_evaluation(f_spaces(m_oa_token_positions[l_token_index- 1]- 1)+ '^ ***');
          trace_evaluation(p_error);

          if l_first_error_token_index= -1
            then l_first_error_token_index:= l_token_index- 1;
          Inc(l_error_count);
*)
        end; // display_evaluation_error

      function f_evaluate_expression: Boolean;

        function f_evaluate_term: Boolean;

          function f_evaluate_factor: Boolean;
            var l_identifier: String;
            begin
              trace_evaluation('> factor ix= '+ IntToStr(l_token_index)+ ' val= '+ f_display_token_type(l_symbol_type));

              case l_symbol_type of
                k_token_opening_parenthesis :
                  begin
                    trace_evaluation('parent');
                    // -- skip (
                    read_next_symbol;

                    Result:= f_evaluate_expression;

                    if l_symbol_type= k_token_closing_parenthesis
                      then read_next_symbol
                      else display_evaluation_error('in "(expr)" manque )');
                  end;

                k_token_NOT :
                  begin
                    trace_evaluation('NOT');
                    read_next_symbol;

                    Result:= NOT f_evaluate_factor;
                    // -- do NOT read next symbol: was read by compile_expression
                  end;

                else
                  if l_symbol_type>= 0
                    then begin
                        trace_evaluation('get id '+ IntToStr(l_symbol_type));

                        l_identifier:= m_c_request_identifiers[l_symbol_type];
                        trace_evaluation('  '+ l_identifier);

                        // -- look if the symbol is in the text
                        Result:= p_c_text_searcher.f_found_string(l_identifier);
                        read_next_symbol;
                      end
                    else begin
                        // -- avoid warnings
                        Result:= False;
                        display_evaluation_error('attend fact: k, var, (: '+ IntToStr(l_symbol_type));
                      end;
              end; // case

              trace_evaluation('< factor ');
            end; // f_evaluate_factor

          begin // f_evaluate_term
            trace_evaluation('> term '+ IntToStr(l_token_index));

            Result:= f_evaluate_factor;

            while l_symbol_type= k_token_AND do
            begin
              read_next_symbol;
              Result:= Result AND f_evaluate_factor;
            end; // while l_symbol_type

            trace_evaluation('< term ');
          end; // f_evaluate_term

        begin // f_evaluate_expression
          trace_evaluation('> expression '+ IntToStr(l_token_index));

          Result:= f_evaluate_term;

          while l_symbol_type= k_token_OR do
          begin
            read_next_symbol;

            Result:= Result OR f_evaluate_term;
          end; // while

          trace_evaluation('< expression ');
        end; // f_evaluate_expression

      begin // f_evaluate_request
        // l_error_count:= 0;
        // l_first_error_token_index:= -1;

        l_token_index:= 0;
        read_next_symbol;

        Result:= f_evaluate_expression;
      end; // f_evaluate_request

    Destructor c_request_evaluator.Destroy;
      begin
        m_c_request_identifiers.Free;

        Inherited;
      end; // Destroy

    begin // u_c_request_evaluator
    end. // u_c_request_evaluator


