// 002 u_c_cgi_site_searcher
// 02 jan 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_cgi_site_searcher;
  interface
    uses Classes
        , u_c_cgi_basic_object
        , u_c_cgi_text_searcher
        , u_c_cgi_request_evaluator

        , u_c_cgi_result_page_list
        ;

    type c_site_searcher= class(c_basic_object)
                            m_c_text_searcher: c_text_searcher;
                            m_c_request_evaluator: c_request_evaluator;
                            m_case_sensitive: Boolean;
                            m_search_root: String;

                            // -- debug
                            m_total_html_page_count: Integer;
                            m_c_file_list: tStringList;
                            m_c_ok_file_list: tStringList;

                            // -- the results are saved here
                            m_c_result_page_list_ref: c_result_page_list;

                            Constructor create_site_searcher(p_name: String);

                              procedure _handle_page(p_path, p_file_name: String);
                            procedure find_pages(p_path: String; p_case_sensitive: Boolean;
                                  p_c_result_page_list: c_result_page_list);
                            procedure search_the_site(p_path, p_request: String;
                                p_case_sensitive: Boolean);

                            Destructor Destroy; Override;
                          end; // c_site_searcher


  implementation
    uses SysUtils
        , u_c_cgi_log
        , u_cgi_strings
        , u_cgi_dir, u_cgi_file
        ;

    // -- c_site_searcher

    Constructor c_site_searcher.create_site_searcher(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_c_file_list:= tStringList.Create;
        m_c_text_searcher:= c_text_searcher.create_text_searcher('search');
        m_c_request_evaluator:= c_request_evaluator.create_request_evaluator('eval');
        m_c_ok_file_list:= tStringList.Create;
      end; // create_site_searcher

    procedure c_site_searcher._handle_page(p_path, p_file_name: String);
        // -- here has a page to analyze
      const k_content_meta= '<META name="DESCRIPTION" content="';
      var l_text_to_search: String;
          l_text_length: Integer;

      function f_extract_content: String;
        var l_start_content, l_index: Integer;
            l_content_meta: String;
        begin
          // -- |<META name="DESCRIPTION" content="last technical papers, ">|
          Result:= '';
          l_content_meta:= k_content_meta;
          if not m_case_sensitive
            then l_content_meta:= LowerCase(l_content_meta);

          l_start_content:= Pos(l_content_meta, l_text_to_search);
          if l_start_content> 0
            then begin
                // -- skip the meta tag
                l_start_content:= l_start_content+ Length(k_content_meta);
                l_index:= l_start_content;

                while (l_index<= l_text_length) and (l_text_to_search[l_index]<> '"') do
                begin
                  Result:= Result+ l_text_to_search[l_index];
                  Inc(l_index);
                end;

                // write_log('content _len='+ IntToStr(Length(Result)));
              end
            else write_log('no_description_meta');
        end; // f_extract_content

      var l_date_time: tDateTime;
          l_c_result_page: c_result_page;

      begin // _handle_page
        m_c_file_list.Add(p_path+ p_file_name);
        Inc(m_total_html_page_count);

        with tStringList.Create do
        begin
          LoadFromFile(p_path+ p_file_name);
          l_text_to_search:= Text;

          l_text_length:= Length(l_text_to_search);
          If l_text_length= 0
            then Exit;

          if not m_case_sensitive
            then l_text_to_search:= LowerCase(l_text_to_search);

          m_c_text_searcher.initialize_text(l_text_to_search);

          with m_c_request_evaluator do
          begin
            if f_evaluate_request(m_c_text_searcher)
              then begin
                  m_c_ok_file_list.Add(p_file_name);

                  // -- fetch the date
                  if m_c_result_page_list_ref<> Nil
                    then begin
                        l_date_time:= f_file_date_time(p_path+ p_file_name);

                        l_c_result_page:= m_c_result_page_list_ref.f_c_add_result_page('page',
                            p_path, p_file_name, l_text_length, l_date_time, f_extract_content);

                        // -- debug
                        // write_log('  p='+ p_path+ '<');
                        // write_log('  f='+ p_file_name+ '<');
                      end;
                end;
          end; // with m_c_request_evaluator

          Free;
        end; // with tStringList
      end; // _handle_page

    procedure c_site_searcher.find_pages(p_path: String; p_case_sensitive: Boolean;
        p_c_result_page_list: c_result_page_list);

      procedure search_pages_recursive(p_level: Integer; p_path: String);
        var l_path: String;
            l_entry: tSearchRec;
            l_result: Integer;
            l_extension: String;
        begin
          l_path:= f_without_ending_slash(p_path);

          // -- find the first entry and save it in l_entry
          l_result:= FindFirst(l_path+ '\*.*', faAnyfile, l_entry);

          if l_result= 0
            then begin
                // -- skip . et .. at the beginning of every path
                if l_entry.Name= '.'
                  then l_result:= FindNext(l_entry);
                if (l_result= 0) and (l_entry.Name= '..')
                  then l_result:= FindNext(l_entry);

                while l_result= 0 do
                begin
                  if (l_entry.Attr and faDirectory)<> 0
                    then begin
                        // -- recurse
                        search_pages_recursive(p_level+ 1, p_path+ '\'+ l_entry.Name);
                      end
                    else begin
                        with l_entry do
                        begin
                          l_extension:= ExtractFileExt(Name);

                          if (LowerCase(l_extension)= '.html')
                              and (Pos('_top.', Name)<= 0) and (Pos('_frame.', Name)<= 0)
                            then _handle_page(p_path+ '\', Name);
                        end;
                      end;

                  // -- the next file of this sub dir
                  l_result:= findnext(l_entry);
                end; // while not errors
              end; // found first
        end; // search_pages_recursive

      var l_path: String;

      begin // find_pages
        m_case_sensitive:= p_case_sensitive;
        m_c_result_page_list_ref:= p_c_result_page_list;
        m_search_root:= p_path;
        l_path:= f_without_ending_slash(p_path);
        m_total_html_page_count:= 0;

        // write_log('  search='+ l_path+ '<');

        search_pages_recursive(0, l_path);

        if m_c_result_page_list_ref<> Nil
          then m_c_result_page_list_ref.m_total_page_count:= m_total_html_page_count;
      end; // find_pages

    procedure c_site_searcher.search_the_site(p_path, p_request: String;
        p_case_sensitive: Boolean);
        // -- for debug
      var l_request: String;
      begin
        l_request:= p_request;
        if not m_case_sensitive
          then l_request:= LowerCase(l_request);

        if m_c_request_evaluator.f_tokenize_request(l_request)
          then begin
              // -- check the syntax
              if m_c_request_evaluator.f_check_syntax
                then begin
                    find_pages(p_path, p_case_sensitive, Nil);
                  end
                else begin
                  end;
            end
          else begin
            end;
      end; // search_the_site

    Destructor c_site_searcher.Destroy;
      begin
        m_c_text_searcher.Free;
        m_c_request_evaluator.Free;
        m_c_file_list.Free;
        m_c_ok_file_list.Free;

        Inherited;
      end; // Destroy


    begin // u_c_yyy
    end. // u_c_yyy
