// 002 u_cgi_coliget_search
// 01 jan 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_cgi_coliget_search;
  Interface

    procedure do_search(p_write_segment,
        p_log_name, p_form_search_root_log_prefix_name,
        p_site_url: String);

    const // -- default to get string to system
          k_output= '';
    var // -- will be filled by the caller in debug mode
        g_output_name: String= k_output;
        g_debug_request: String= '';
        g_debug_search_path: String= '';
        g_case_sensitive: Boolean= True;

  Implementation
    uses Classes, SysUtils
        , u_c_cgi_log, u_cgi_helpers
        , u_c_cgi_forms

        , u_c_cgi_basic_object
        , u_cgi_strings
        , u_cgi_strings_html

        , u_c_cgi_site_searcher
        , u_c_cgi_result_page_list
        , u_c_cgi_request_evaluator
        , u_c_cgi_html_result_page_builder
        ;

    const k_content_type= 'Content-type: text/html';

    procedure do_search(p_write_segment,
        p_log_name, p_form_search_root_log_prefix_name,
        p_site_url: String);

      function f_is_debug: Boolean;
        begin
          Result:= g_output_name<> k_output;
        end; // f_is_debug

      procedure analyze_form_values(var pv_request: String; var pv_case_sensitive: Boolean);

        function f_convert_cgi_question_string(p_cgi_string: String): String;
           // -- convert "q= xxx&"
           // -- convert "q=uml&minuscule=on&
          var l_index, l_length: Integer;
              l_hex_string: String[3];
              l_special_ascii_code: Byte;
          begin
            // -- "q=midas+AND+dbase&"
            l_length:= Length(p_cgi_string);

            if l_length= 0
              then Result:= ''
              else
                // -- starts with "q=" and terminates with "&"
                if (l_length>= 3) and (p_cgi_string[1]='q') and (p_cgi_string[2]='=')
                    and (p_cgi_string[l_length]='&')
                  then begin
                      // Result:= Copy(p_cgi_string, 3, l_length- 3);
                      Result:= '';
                      // -- skip the starting "q="
                      l_index:= 3;

                      // -- skip the ending "&"
                      Dec(l_length, 1);
                      l_hex_string:= '$  ';

                      while l_index<= l_length do
                      begin
                        Case p_cgi_string[l_index] of
                          '+' : Result:= Result+ ' ';
                          '%' : begin
                                  // -- skip %
                                  Inc(l_index);
                                  l_hex_string[2]:= p_cgi_string[l_index];
                                  Inc(l_index);
                                  l_hex_string[3]:= p_cgi_string[l_index];

                                  l_special_ascii_code:= StrToInt(l_hex_string);
                                  Result:= Result+ chr(l_special_ascii_code);
                                end;
                          '&' : result:= Result+ ' ';
                          else
                            Result:= Result+ p_cgi_string[l_index];
                        end; // case

                        Inc(l_index);
                      end; // while
                    end // q=xxx&
                  else Result:= '';
          end; // f_convert_cgi_question_string

        var l_raw_request_string: String;
            l_minuscule_position: Integer;
            l_minuscule_string: String;

        begin // analyze_form_values
          with c_form.create_form('forms', f_cgi_exe_path
              + p_write_segment+ p_form_search_root_log_prefix_name) do
          begin
            get_form_result;

            l_raw_request_string:= m_raw_answer_string;
            pv_request:= f_convert_cgi_question_string(l_raw_request_string);

            l_minuscule_position:= Pos('minuscule', pv_request);
            if l_minuscule_position> 0
              then begin
                  // -- the end of the string
                  l_minuscule_string:= Copy(pv_request, l_minuscule_position,
                      Length(pv_request)+ 1- l_minuscule_position);
                  pv_case_sensitive:= Pos('=on', l_minuscule_string)<= 0;
                  pv_request:= Trim(Copy(pv_request, 1, l_minuscule_position- 1));
                end
              else pv_case_sensitive:= True;

            write_log('Received:');
            write_log('  Rem >'+ m_remote_address_string+ '<');
            write_log('  raw >'+ l_raw_request_string+ '<');
            write_log('  req >'+ pv_request+ '<');
            write_log('  min >'+ l_minuscule_string+ '<');
            if pv_case_sensitive
              then write_log('  case_sensitive')
              else write_log('  case_insensitive');
            // write_log('  dir >'+ f_cgi_exe_path+ '<');

            Free;
          end; // with c_form.create_form
        end; // analyze_form_values

      var l_c_html_result_page_builder: c_html_result_page_builder;

      procedure evaluate_request(p_request: String);

        procedure log_request_error(p_c_request_evaluator: c_request_evaluator);
            // -- debug
          var l_index: Integer;
          begin
            with p_c_request_evaluator do
            begin
              write_log('syntax_err '+ Format('err_ix= %3d len= %3d tok=%3d typ= ',
                  [m_error_index, m_error_length, m_error_token_index])
                  + f_error_message(m_error_type)+ ' '+ m_error_message+ '<');
              for l_index:= 0 to m_token_count- 1 do
              begin
                write_log(Format('ix= %2d  tok= %3d   err=%3d  len=%3d',
                  [l_index, m_oa_tokens[l_index], m_oa_token_positions[l_index], m_oa_token_lengths[l_index]]));
              end;
            end; // with p_c_request_evaluator
          end; // log_request_error

        var l_c_result_page_list: c_result_page_list;
            l_search_path: String;

        begin // evaluate_request
          write_log('  evaluate >'+ p_request+ '<');

          with c_site_searcher.create_site_searcher('search') do
          begin
            if m_c_request_evaluator.f_tokenize_request(p_request)
              then begin
                  // -- check the syntax
                  if m_c_request_evaluator.f_check_syntax
                    then begin
                        l_c_result_page_list:= c_result_page_list.create_result_page_list('page_list');

                        if f_is_debug
                          then l_search_path:= g_debug_search_path
                          else begin
                              l_search_path:= f_cgi_exe_path;
                              // -- remove scripts
                              l_search_path:= f_remove_end_if_end_is_equal_to(Lowercase(l_search_path), 'scripts\');
                            end;

                        // write_log('search_path '+ l_search_path);

                        find_pages(l_search_path, g_case_sensitive, l_c_result_page_list);

                        // -- now use the page list to build the answer
                        write_log('  build_result_pages');
                        l_c_html_result_page_builder.build_the_result_page(l_c_result_page_list);

                        l_c_result_page_list.Free;
                      end
                    else begin
                        // -- build the syntax error answer
                        with m_c_request_evaluator do
                        begin
                          log_request_error(f_c_self);

                          l_c_html_result_page_builder.build_error_page(m_error_index,
                              m_error_length, m_error_message);
                        end;
                      end;
                end
              else begin
                  // -- build the lexical error answer
                  with m_c_request_evaluator do
                    l_c_html_result_page_builder.build_error_page(m_error_index,
                        m_error_length, m_error_message);
                end;

            Free;
          end; // with c_site_searcher
        end; // evaluate_request

      procedure send_answer(p_output_name: String);
        var l_output: Text;

        procedure send_empty_answer;
          begin
            writeln(l_output, '<HTML>');
            writeln(l_output, '  <BODY>');
            writeln(l_output, '    Pas de rponse');
            writeln(l_output, '  </BODY>');
            writeln(l_output, '</HTML>');
          end; // send_empty_answer

        var l_line_index: Integer;

        begin // send_answer
          // -- now build the answer
          Assign(l_output, p_output_name);
          Rewrite(l_output);

          // writeln(l_output, 'HTTP/1.0 200 OK');
          if not f_is_debug
            then begin
                writeln(l_output, k_content_type);
                writeln(l_output);
              end;

          write_log('  => Answser: header_ok');

          // send_empty_answer;

          with l_c_html_result_page_builder.m_c_html_page do
            if Count= 0
              then send_empty_answer
              else
                for l_line_index:= 0 to Count- 1 do
                  writeln(l_output, Strings[l_line_index]);

          Close(l_output);

          write_log('  ok_sent');
        end; // send_answer

      procedure answer_ok;
          // -- bogus trial
        var l_output: Text;
        begin
          // -- now build the answer
          Assign(l_output, g_output_name);
          Rewrite(l_output);

          // writeln(l_output, 'HTTP/1.0 200 OK');
          writeln(l_output, k_content_type);
          writeln(l_output);

          writeln(l_output, '<HTML>');
          writeln(l_output, '  <BODY>');
          writeln(l_output, '    log_ok ok');
          writeln(l_output, '  </BODY>');
          writeln(l_output, '</HTML>');

          Close(l_output);
        end; // answer_ok

      var l_request: String;
          l_site_url: String;

      begin // do_search
        open_log(f_cgi_exe_path+ p_write_segment+ p_log_name);

        // -- a bogus answer for debugging
        // answer_ok;

        // -- get the request without % etc
        write_log('== analyze_form_values');
        if f_is_debug
          then l_request:= g_debug_request
          else analyze_form_values(l_request, g_case_sensitive);

        if not g_case_sensitive
          then l_request:= LowerCase(l_request);
        write_log('  req+case '+ l_request);

        if f_is_debug
          then l_site_url:= ''
          else l_site_url:= p_site_url;

        l_c_html_result_page_builder:=
            c_html_result_page_builder.create_html_result_page_builder(l_request,
            l_site_url);

        // -- build the target file list
        evaluate_request(l_request);

        write_log('== send_answer');
        send_answer(g_output_name);
        // -- in debug mode, also output an .html version of the answer on the local disc
        if f_is_debug
          then send_answer(ChangeFileExt(g_output_name, '.html'));

        l_c_html_result_page_builder.Free;
        write_log('== end ok');
      end; // do_search

    end.